# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts

test_data_folder = xt._pkg_root.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_gaussian_bunch_generation(test_context):
    bunch_intensity = 1e11
    sigma_z = 22.5e-2
    n_part = int(5e6)
    nemitt_x = 2e-6
    nemitt_y = 2.5e-6

    filename = test_data_folder.joinpath(
        'sps_w_spacecharge/line_no_spacecharge_and_particle.json')
    with open(filename, 'r') as fid:
        ddd = json.load(fid)
    line = xt.Line.from_dict(ddd['line'])
    line.particle_ref = xp.Particles.from_dict(ddd['particle'],
                           _context=test_context # for testing purposes
                           )

    line.build_tracker(_context=test_context)

    part_on_co = line.find_closed_orbit()

    part = xp.generate_matched_gaussian_bunch(
             _context=test_context,
             line=line,
             num_particles=n_part,
             total_intensity_particles=bunch_intensity,
             nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z)

    tw = line.twiss()

    # CHECKS
    y_rms = np.std(test_context.nparray_from_context_array(part.y))
    py_rms = np.std(test_context.nparray_from_context_array(part.py))
    x_rms = np.std(test_context.nparray_from_context_array(part.x))
    px_rms = np.std(test_context.nparray_from_context_array(part.px))
    delta_rms = np.std(test_context.nparray_from_context_array(part.delta))
    zeta_rms = np.std(test_context.nparray_from_context_array(part.zeta))

    part_on_co.move(_context=xo.ContextCpu())
    gemitt_x = nemitt_x/part_on_co.beta0/part_on_co.gamma0
    gemitt_y = nemitt_y/part_on_co.beta0/part_on_co.gamma0
    assert np.isclose(zeta_rms, sigma_z, rtol=1e-2, atol=1e-15)
    assert np.isclose(x_rms,
                      np.sqrt(tw['betx'][0]*gemitt_x + tw['dx'][0]**2*delta_rms**2),
                      rtol=1e-2, atol=1e-15)
    assert np.isclose(y_rms,
                      np.sqrt(tw['bety'][0]*gemitt_y + tw['dy'][0]**2*delta_rms**2),
                      rtol=1e-2, atol=1e-15)


@for_all_test_contexts
def test_short_bunch(test_context):
    bunch_intensity = 1e11
    sigma_z = 22.5e-2/5
    n_part = int(5e3)
    nemitt_x = 2e-6
    nemitt_y = 2.5e-6

    filename = test_data_folder.joinpath(
            'sps_w_spacecharge/line_no_spacecharge_and_particle.json')
    with open(filename, 'r') as fid:
        ddd = json.load(fid)
    line = xt.Line.from_dict(ddd['line'])
    particle_ref = xp.Particles.from_dict(ddd['particle'])
    line.build_tracker(_context=test_context)

    tw = line.twiss(particle_ref=particle_ref)

    part = xp.generate_matched_gaussian_bunch(
             _context=test_context,
             num_particles=n_part, total_intensity_particles=bunch_intensity,
             nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
             particle_ref=particle_ref, line=line)

    # CHECKS
    y_rms = np.std(test_context.nparray_from_context_array(part.y))
    py_rms = np.std(test_context.nparray_from_context_array(part.py))
    x_rms = np.std(test_context.nparray_from_context_array(part.x))
    px_rms = np.std(test_context.nparray_from_context_array(part.px))
    delta_rms = np.std(test_context.nparray_from_context_array(part.delta))
    zeta_rms = np.std(test_context.nparray_from_context_array(part.zeta))

    gemitt_x = nemitt_x/particle_ref.beta0[0]/particle_ref.gamma0[0]
    gemitt_y = nemitt_y/particle_ref.beta0[0]/particle_ref.gamma0[0]
    assert np.isclose(zeta_rms, sigma_z, rtol=5e-2, atol=1e-15)
    assert np.isclose(x_rms,
                      np.sqrt(tw['betx'][0]*gemitt_x + tw['dx'][0]**2*delta_rms**2),
                      rtol=5e-2, atol=1e-15)
    assert np.isclose(y_rms,
                      np.sqrt(tw['bety'][0]*gemitt_y + tw['dy'][0]**2*delta_rms**2),
                      rtol=5e-2, atol=1e-15)

    for iturn in range(100):
        line.track(part)
        assert np.isclose(zeta_rms, sigma_z, rtol=5e-2, atol=1e-15)



import random

import pytest

from xpart.test_helpers import flaky_assertions, retry


@retry(on=AssertionError, n_times=100)
def test_retry_passes():
    """Below assertion almost always fails, however the test should
    practically always pass (1 in 38k chance of failure)."""
    assert random.random() > 0.9


def test_retry_fails():
    @retry(on=AssertionError, n_times=2)
    def actual_test():
        """Below assertion practically always fails, thus the test is
        expected to fail as well after only 2 runs."""
        assert random.random() > 0.99_999

    with pytest.raises(AssertionError):
        actual_test()


@pytest.mark.parametrize(
    'inner_when, outer_when, error_substring',
    [
        (3, 3, 'inner'),
        (0, 3, 'outer'),
        (3, 0, 'inner'),
        (3, 1, 'inner'),
        (2, 1, 'outer'),
        (0, 0, None),
        (2, 0, None),
    ]
)
def test_retry_with_context_many_errors_fails(
        inner_when,
        outer_when,
        error_substring
):
    class FlakyFailer:
        """A class simulating flaky behaviour.

        The failer shall succeed after `succeed_after` failures with an
        assertion error containing `name` in the message.
        """
        def __init__(self, succeed_after, name):
            self.name = name
            self.when = succeed_after

        def test(self):
            if self.when == 0:
                return

            self.when -= 1
            raise AssertionError(f'Failing {self.name}')

    inner_failer = FlakyFailer(succeed_after=inner_when, name='inner')
    outer_failer = FlakyFailer(succeed_after=outer_when, name='outer')

    @retry(n_times=3)
    def actual_test():
        with flaky_assertions():
            inner_failer.test()

        outer_failer.test()

    if error_substring:  # if `error_substring` is None, assert test failure
        with pytest.raises(AssertionError) as e_info:
            actual_test()

        assert error_substring in e_info.exconly()
    else:  # assert test success
        actual_test()
